import logging
from abc import abstractmethod
from dataclasses import dataclass
from functools import partial
from json.decoder import JSONDecodeError
from typing import Union

import requests

from pydispix.client import Client
from pydispix.color import Color, parse_color
from pydispix.errors import RateLimitBreached
from pydispix.utils import resolve_url_endpoint

logger = logging.getLogger("pydispix")


@dataclass
class ChurchTask:
    x: int
    y: int
    color: Union[int, str, tuple[int, int, int], Color]


class ChurchClient(Client):
    def __init__(
        self,
        pixel_api_token: str,
        church_token: str,
        base_church_url: str,
        *args,
        **kwargs
    ):
        super().__init__(pixel_api_token, *args, **kwargs)

        if not base_church_url.endswith("/"):
            base_church_url = base_church_url + "/"

        self.base_church_url = base_church_url
        self.church_token = church_token

    def resolve_church_endpoint(self, endpoint: str):
        return resolve_url_endpoint(self.base_church_url, endpoint)

    @abstractmethod
    def get_task(self, endpoint: str = "get_task", repeat_delay: int = 5) -> ChurchTask:
        """
        Get task from the church, this is an abstract method, you'll need
        to override this to get it to work with your church's specific API.

        `repeat_delay` is the time we will wait for, if the church currently
        doesn't have any aviable tasks for us.
        """

    @abstractmethod
    def submit_task(self, church_task: ChurchTask, endpoint: str = "submit_task") -> requests.Response:
        """
        Submit a task to the church, this is an abstract method, you'll need
        to override this to get it to work with your church's specific API.
        """

    def run_task(
        self,
        submit_endpoint: str = "submit_task",
        show_progress: bool = False,
        repeat_delay: int = 5,
        repeat_on_ratelimit: bool = True,
    ):
        """
        Obtain the Church Task, put new pixel on the canvas and send the `submit_task` request.

        In case we hit initial rate limit from `set_pixel`, the rate limit is waited out,
        and we proceed with a new task, since the church's API time limit for that task
        has already likely expired.

        `repeat_delay` is the delay to wait, if the church doesn't have any more aviable tasks.
        """
        # This can't just use the `set_pixel`, because we need to send submit message to the church
        # before we wait for the rate limits, this is also why we use `make_raw_request` instead
        # of just using `make_requests` that handles the rate limits for us
        task = self.get_task(repeat_delay=repeat_delay)
        logger.info(f"Running church task: {task}")

        # Manual set_pixel, with submit before waiting for rate limits
        url = self.resolve_endpoint("set_pixel")
        try:
            response = self.make_request(
                "POST", url,
                data={
                    "x": task.x,
                    "y": task.y,
                    "rgb": parse_color(task.color)
                },
                ratelimit_after=True,
                task_after=partial(self.submit_task, task, endpoint=submit_endpoint),
                show_progress=show_progress
            )
        except RateLimitBreached as exc:
            try:
                response_text = exc.response.json()["message"]
            except (JSONDecodeError, KeyError):
                response_text = exc.response.content

            # This can occur first time we run this, since we're handling
            # rate limits after the request is made, `repeat_on_ratelimit`
            # should always be `True` initially. We can then wait out the
            # rate limit, and re-run the whole function with a new task,
            # since the time limit for the completion of this one has most
            # likely already expired.
            if repeat_on_ratelimit:
                logger.warning(f"Hit pixels api ratelimit: {response_text}, waiting it out and ignoring this task.")
                self.rate_limiter.wait(url, show_progress=show_progress)
                # Re-run the task only once, this rate breach should only occur
                # on initial request, if it happens again, it shouldn't be handled
                return self.run_task(
                    submit_endpoint=submit_endpoint,
                    show_progress=show_progress,
                    repeat_delay=repeat_delay,
                    repeat_on_ratelimit=False
                )
            raise exc

        # Return status of the submit task, or raise the exception that ocurred in it
        if hasattr(response, "task_exception"):
            raise response.task_exception
        return response.task_result

    def run_tasks(
        self,
        submit_endpoint: str = "submit_task",
        show_progress: bool = False,
        repeat_delay: int = 5
    ):
        """Keep running church tasks forever."""
        while True:
            self.run_task(
                submit_endpoint=submit_endpoint,
                show_progress=show_progress,
                repeat_delay=repeat_delay
            )



import logging
import random
import re
import time
from dataclasses import dataclass
from json.decoder import JSONDecodeError

import requests

from pydispix.church import ChurchClient, ChurchTask
from pydispix.errors import RateLimitBreached

logger = logging.getLogger("pydispix")

SQLITE_CHURCH = "https://decorator-factory.su"
RICK_CHURCH = "https://pixel-tasks.scoder12.repl.co/api"


@dataclass
class RickChurchTask(ChurchTask):
    project_title: str
    start: float


@dataclass
class SQLiteChurchTask(ChurchTask):
    id: int
    issued_by: str


class RickChurchClient(ChurchClient):
    """Church Client designed to work specifically with rick church"""
    def __init__(
        self,
        pixel_api_token: str,
        church_token: str,
        base_church_url: str = RICK_CHURCH,
        *args,
        **kwargs
    ):
        super().__init__(pixel_api_token, church_token, base_church_url, *args, **kwargs)

    @property
    def personal_stats(self):
        """Get personal stats."""
        url = self.resolve_church_endpoint("user/stats")
        return self.make_request("GET", url, params={"key": self.church_token}).json()

    @property
    def church_stats(self):
        """Get church stats."""
        url = self.resolve_church_endpoint("overall_stats")
        return self.make_request("GET", url).json()

    @property
    def leaderboard(self) -> list:
        """Get church leaderboard."""
        url = self.resolve_church_endpoint("leaderboard")
        return self.make_request("GET", url).json()["leaderboard"]

    @property
    def uptime(self) -> float:
        """Uptime of the church of rick."""
        url = self.resolve_church_endpoint("leaderboard")
        return float(self.make_request("GET", url).json()["uptime"])

    @property
    def projects(self) -> list:
        """Get project data from the church."""
        url = self.resolve_church_endpoint("projects/stats")
        return self.make_request("GET", url).json()

    def get_task(self, repeat_delay: int = 5) -> RickChurchTask:
        url = self.resolve_church_endpoint("get_task")
        while True:
            response = self.make_request("GET", url, params={"key": self.church_token}).json()

            if response["task"] is None:
                logger.info(f"Church doesn't currently have any aviable tasks, waiting {repeat_delay}s")
                time.sleep(repeat_delay)
                continue
            return RickChurchTask(**response["task"])

    def submit_task(self, church_task: RickChurchTask, endpoint: str = "submit_task") -> requests.Response:
        url = self.resolve_church_endpoint(endpoint)
        body = {
            'project_title': church_task.project_title,
            'start': church_task.start,
            'x': church_task.x,
            'y': church_task.y,
            'color': church_task.color
        }
        req = self.make_request("POST", url, data=body, params={"key": self.church_token})
        logger.info(f"Task submitted to the church (tasks complete={self.personal_stats['goodTasks']})")
        return req

    def run_task(
        self,
        submit_endpoint: str = "submit_task",
        show_progress: bool = False,
        repeat_delay: int = 5,
        repeat_on_ratelimit: bool = True,
    ) -> requests.Response:
        """
        Extend `run_task` to handle specific exceptions from the
        church of rick.
        """
        try:
            return super().run_task(
                submit_endpoint=submit_endpoint,
                show_progress=show_progress,
                repeat_delay=repeat_delay,
                repeat_on_ratelimit=repeat_on_ratelimit,
            )
        except RateLimitBreached as exc:
            # If we take longer to submit a request to the church, it will
            # result in RateLimitBreached
            try:
                details = exc.response.json()["detail"]
            except (JSONDecodeError, KeyError):
                # If response isn't json decodeable or doesn't contain a detail key,
                # it isn't from rick church
                raise exc
            if not re.search(
                r"You have not gotten a task yet or you took more than \d+ seconds to submit your task",
                details
            ):
                # If we didn't catch this error message, something else has ocurred
                # or this wasn't an exception from the rick church, don't handle it
                raise exc
            logger.warn("Church task failed, got rate limited: submitting task took too long")
            return self.run_task(
                submit_endpoint=submit_endpoint,
                show_progress=show_progress,
                repeat_on_ratelimit=repeat_on_ratelimit
            )


class SQLiteChurchClient(ChurchClient):
    """Church Client designed to work specifically with rick church"""
    def __init__(
            self,
            pixel_api_token: str,
            base_church_url: str = SQLITE_CHURCH,
            *args,
            **kwargs
    ):
        # SQLite Church API is open for everyone, it doesn't need a token
        church_token = None
        super().__init__(pixel_api_token, church_token, base_church_url, *args, **kwargs)

    def get_task(self, endpoint: str = "tasks", repeat_delay: int = 5) -> SQLiteChurchTask:
        url = self.resolve_church_endpoint(endpoint)
        while True:
            response = self.make_request("GET", url)

            if len(response) == 0:
                logger.info(f"Church doesn't currently have any aviable tasks, waiting {repeat_delay}s")
                time.sleep(repeat_delay)
                continue
            # SQLite church returns a list of aviable tasks to complete, it doesn't assign
            # specific tasks to members, since there is no unique API key. Best we can do is
            # Therefore to pick a task randomly from this list
            task = random.choice(response)
            return SQLiteChurchTask(**task)

    def submit_task(self, church_task: SQLiteChurchTask, endpoint: str = "submit_task") -> requests.Response:
        url = self.resolve_church_endpoint(endpoint)
        body = {"task_id": church_task.id}
        req = self.make_request("POST", url, data=body)
        logger.info("Task submitted to the church")
        return req



import os
import requests
import logging
from collections import namedtuple
from typing import Callable, Union, Optional

from pydispix.ratelimits import RateLimiter
from pydispix.canvas import Canvas, Pixel
from pydispix.color import Color, parse_color
from pydispix.errors import RateLimitBreached, InvalidToken, handle_invalid_body
from pydispix.utils import resolve_url_endpoint

logger = logging.getLogger("pydispix")
Dimensions = namedtuple("Dimensions", ("width", "height"))


class Client:
    """HTTP client to the pixel API."""
    def __init__(self, token: Optional[str] = None, base_url: str = "https://pixels.pythondiscord.com/"):
        if token is None:
            try:
                token = os.environ["TOKEN"]
            except KeyError:
                raise RuntimeError("Unable to load token, 'TOKEN' environmental variable not found.")

        if not base_url.endswith("/"):
            base_url = base_url + "/"

        self.token = token
        self.base_url = base_url
        self.headers = {
            "Authorization": "Bearer " + token,
            "User-Agent": "ItsDrike pydispix",
        }
        self.rate_limiter = RateLimiter()

    def make_raw_request(
        self, method: str, url: str, *,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        update_rate_limits: bool = True,
    ) -> requests.Response:
        """
        This method is here purely to make an HTTP request and update the rate limiter.
        Even though this will update the rate limtis, it will not wait for them.
        """
        logger.debug(f"Request: {method} on {url} {data=} {params=}.")
        response = requests.request(
            method, url,
            json=data,
            params=params,
            headers=self.headers
        )

        if update_rate_limits:
            self.rate_limiter.update_from_headers(url, response.headers)

        if response.status_code == 429:
            logger.debug(f"Request failed (rate limitation): {method} on {url} {data=} {params=}")
            raise RateLimitBreached(
                "Request didn't succeed because it was made during a rate-limit phase.",
                response=response
            )
        if response.status_code == 401:
            logger.error("Request failed with 401 (Forbidden) code. This means your API token is most likely invalid.")
            raise InvalidToken("Received 401 - FORBIDDEN: Is your API token correct?", response=response)

        if response.status_code == 422:
            exc = handle_invalid_body(response)
            if exc is not None:
                raise exc

        if response.status_code != 200:
            raise requests.HTTPError(f"Received code {response.status_code}", response=response)

        return response

    def make_request(
        self, method: str, url: str, *,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        ratelimit_after: bool = False,
        task_after: Optional[Callable] = None,
        head_ratelimit_update: bool = False,
        repeat_on_ratelimit: bool = False,
        show_progress: bool = False,
    ) -> requests.Response:
        """
        This method handles making a request on a rate-limited endpoint.

        `ratelimit_after`: Wait for rate limits after making the request, this is handy
        for some requests, which should be ran as soon as possible, and wait out the rate
        limit after it was made.

        `task_after`: This is a function that will run after the request was made. This is
        useful when reporting that some task was fullfilled to external API (like to a church)
        where we can't afford to wait out the rate limit first. This is usually combined with
        ratelimit_after`, since there's no point in using it if we wait out the rate limit first.
        The result of this task will be stored in the response under `task_result` variable.
        In case any exception would occur in this task, it will be handled and stored in
        in the response under `task_exception` variable.

        `head_ratelimit_update`: Some APIs allow sending HEAD type requests, which don't actually
        trigger interfere with the API, and are here purely to obtain the rate limits, so that we
        can wait it out before we make an actual request. This can only be used when `ratelimit_after`
        isn't being used, since if it is, we'll obtain rate-limits from the original made request,
        instead of the HEAD.

        `repeat_on_ratelimit`: This can be used to repeat this request if request gives us
        `RateLimitBreached` exception. This will re-run the whole function again one more time, but
        if the second request fails too, `RateLimitBreached` will be raised anyway. (To avoid infinite
        loops). This option can't be used with `ratelimit_after` since if we breached rate limit, we
        have to wait it out, and we can't wait it out after the request we made has failed.
        """
        if repeat_on_ratelimit and ratelimit_after:
            raise ValueError(
                "Can't combine `ratelimit_after` with `repeat_on_ratelimit` (If we breached rate-limit, "
                "we have to wait it out, and that's impossible to do after a failed request)"
            )

        if not ratelimit_after:
            if head_ratelimit_update:
                self.make_raw_request("HEAD", url, update_rate_limits=True)
            self.rate_limiter.wait(url, show_progress=show_progress)

        try:
            response = self.make_raw_request(
                method, url,
                data=data, params=params,
                update_rate_limits=True
            )
        except RateLimitBreached as exc:
            if repeat_on_ratelimit:
                logger.warning(f"Hit rate limit, repeating request ({exc.response.content})")
                # There's no point in using `head_ratelimit_update` here, since the failed
                # request has already updated the rate limits.
                return self.make_request(
                    method, url,
                    data=data, params=params,
                    ratelimit_after=False,
                    task_after=task_after, head_ratelimit_update=False,
                    repeat_on_ratelimit=False, show_progress=show_progress
                )
            raise exc

        if task_after:
            try:
                result = task_after()
            except Exception as exc:
                response.task_exception = exc
            else:
                response.task_result = result

        if ratelimit_after:
            self.rate_limiter.wait(url, show_progress=show_progress)

        return response

    def resolve_endpoint(self, endpoint: str) -> str:
        """Resolve given `endpoint` to use the base_url"""
        return resolve_url_endpoint(self.base_url, endpoint)

    def get_dimensions(self) -> Dimensions:
        """Make a request to obtain the canvas dimensions"""
        url = self.resolve_endpoint("get_size")
        data = self.make_request("GET", url).json()
        return Dimensions(width=data["width"], height=data["height"])

    def get_canvas(self, show_progress: bool = False) -> Canvas:
        """Fetch the whole canvas and return it in a `Canvas` object."""
        url = self.resolve_endpoint("get_pixels")
        data = self.make_request("GET", url, show_progress=show_progress).content
        size = self.get_dimensions()
        return Canvas(size, data)

    def get_pixel(self, x: int, y: int, show_progress: bool = False) -> Pixel:
        """Fetch rgb data about a specific pixel"""
        url = self.resolve_endpoint("get_pixel")
        data = self.make_request("GET", url, params={"x": x, "y": y}, show_progress=show_progress).json()
        hex_color = data["rgb"]
        return Pixel.from_hex(hex_color)

    def put_pixel(
        self,
        x: int, y: int,
        color: Union[int, str, tuple[int, int, int], Color],
        show_progress: bool = False,
    ) -> str:
        """
        Draw a pixel and return a message.

        If you are reporting to some church, you may want to set
        `task_after` to the completion request for your church.
        This way, the task will be executed instantly, before
        waiting for the rate limitation
        """
        url = self.resolve_endpoint("set_pixel")
        data = self.make_request(
            "POST", url,
            data={
                "x": x,
                "y": y,
                "rgb": parse_color(color)
            },
            head_ratelimit_update=True,
            show_progress=show_progress,
        )

        msg = data.json()["message"]
        logger.info(f"Success: {msg}")
        return msg
